@@ -28,6 +28,7 @@
 #include <fcntl.h>  /* open */
 #include <unistd.h>  /* STDOUT_FILENO */
 #include <signal.h>  /* sigaction */
+#include <stdlib.h>  /* exit */
 #include <termios.h>  /* tcgetattr, tcsetattr */
 
 #include <chafa.h>
@@ -37,6 +38,7 @@
 
 #define ANIM_FPS_MAX 100000.0
 #define FILE_DURATION_DEFAULT 0.0
+#define SCALE_MAX 9999.0
 
 typedef struct
 {
@@ -44,6 +46,7 @@ typedef struct
 
     gboolean show_help;
     gboolean show_version;
+    gboolean skip_processing;
 
     GList *args;
     ChafaCanvasMode mode;
@@ -72,6 +75,7 @@ typedef struct
     gint width, height;
     gint cell_width, cell_height;
     gint margin_bottom, margin_right;
+    gdouble scale;
     gdouble font_ratio;
     gint work_factor;
     gint optimization_level;
@@ -108,6 +112,7 @@ TermSize;
 
 static GlobalOptions options;
 static TermSize detected_term_size;
+static gboolean using_detected_size = FALSE;
 static volatile sig_atomic_t interrupted_by_user = FALSE;
 
 static void
@@ -165,6 +170,20 @@ get_hex_byte (const gchar *str)
     return b;
 }
 
+static gint
+count_dash_strings (GList *l)
+{
+    gint n = 0;
+
+    for ( ; l; l = g_list_next (l))
+    {
+        if (!strcmp (l->data, "-"))
+            n++;
+    }
+
+    return n;
+}
+
 static gboolean
 parse_color (const gchar *str, guint32 *col_out, GError **error)
 {
@@ -232,15 +251,20 @@ print_version (void)
 {
     gchar *builtin_features_str = chafa_describe_features (chafa_get_builtin_features ());
     gchar *supported_features_str = chafa_describe_features (chafa_get_supported_features ());
+    gchar **loaders = get_loader_names ();
+    gchar *loaders_joined = g_strjoinv (" ", loaders);
 
-    g_print ("Chafa version %s\n\nFeatures: %s\nApplying: %s\n\n%s\n",
+    g_print ("Chafa version %s\n\nLoaders:  %s\nFeatures: %s\nApplying: %s\n\n%s\n",
              CHAFA_VERSION,
+             loaders_joined,
              chafa_get_builtin_features () ? builtin_features_str : "none",
              chafa_get_supported_features () ? supported_features_str : "none",
              copyright_notice);
 
     g_free (builtin_features_str);
     g_free (supported_features_str);
+    g_strfreev (loaders);
+    g_free (loaders_joined);
 }
 
 static void
@@ -269,8 +293,8 @@ print_summary (void)
     "      --bg=COLOR     Background color of display (color name or hex).\n"
     "  -C, --center=BOOL  Center images [on, off]. Defaults to off.\n"
     "      --clear        Clear screen before processing each file.\n"
-    "  -c, --colors=MODE  Set output color mode; one of [none, 2, 8, 16, 240, 256,\n"
-    "                     full]. Defaults to full (24-bit).\n"
+    "  -c, --colors=MODE  Set output color mode; one of [none, 2, 8, 16/8, 16, 240,\n"
+    "                     256, full]. Defaults to full (24-bit).\n"
     "      --color-extractor=EXTR  Method for extracting color from an area\n"
     "                     [average, median]. Average is the default.\n"
     "      --color-space=CS  Color space used for quantization; one of [rgb, din99d].\n"
@@ -316,6 +340,10 @@ print_summary (void)
     "                     altered state (rude).\n"
     "  -p, --preprocess=BOOL  Image preprocessing [on, off]. Defaults to on with 16\n"
     "                     colors or lower, off otherwise.\n"
+    "      --scale=NUM    Scale image, respecting terminal's maximum dimensions. 1.0\n"
+    "                     approximates original pixel dimensions. Specify \"max\" to\n"
+    "                     use all available space. Defaults to 1.0 for pixel graphics\n"
+    "                     and 4.0 for symbols.\n"
     "  -s, --size=WxH     Set maximum output dimensions in columns and rows. By\n"
     "                     default this will be the size of your terminal, or 80x25\n"
     "                     if size detection fails.\n"
@@ -323,7 +351,7 @@ print_summary (void)
     "                     either a unitless multiplier, or a real number followed\n"
     "                     by \"fps\" to apply a specific framerate.\n"
     "      --stretch      Stretch image to fit output dimensions; ignore aspect.\n"
-    "                     Implies --zoom.\n"
+    "                     Implies --scale max.\n"
     "      --symbols=SYMS  Specify character symbols to employ in final output.\n"
     "                     See below for full usage and a list of symbol classes.\n"
     "      --threads=NUM  Maximum number of CPU threads to use. If left unspecified\n"
@@ -334,8 +362,7 @@ print_summary (void)
     "                     contents change. Will run until manually interrupted\n"
     "                     or, if --duration is set, until it expires.\n"
     "  -w, --work=NUM     How hard to work in terms of CPU and memory [1-9]. 1 is the\n"
-    "                     cheapest, 9 is the most accurate. Defaults to 5.\n"
-    "      --zoom         Allow scaling up beyond one character per pixel.\n\n"
+    "                     cheapest, 9 is the most accurate. Defaults to 5.\n\n"
 
     "  Accepted classes for --symbols and --fill are [all, none, space, solid,\n"
     "  stipple, block, border, diagonal, dot, quad, half, hhalf, vhalf, inverted,\n"
@@ -372,10 +399,13 @@ parse_colors_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_
         options.mode = CHAFA_CANVAS_MODE_FGBG;
     else if (!g_ascii_strcasecmp (value, "2"))
         options.mode = CHAFA_CANVAS_MODE_FGBG_BGFG;
-    else if (!g_ascii_strcasecmp (value, "16"))
-        options.mode = CHAFA_CANVAS_MODE_INDEXED_16;
     else if (!g_ascii_strcasecmp (value, "8"))
         options.mode = CHAFA_CANVAS_MODE_INDEXED_8;
+    else if (!g_ascii_strcasecmp (value, "16-8")
+             || !g_ascii_strcasecmp (value, "16/8"))
+        options.mode = CHAFA_CANVAS_MODE_INDEXED_16_8;
+    else if (!g_ascii_strcasecmp (value, "16"))
+        options.mode = CHAFA_CANVAS_MODE_INDEXED_16;
     else if (!g_ascii_strcasecmp (value, "240"))
         options.mode = CHAFA_CANVAS_MODE_INDEXED_240;
     else if (!g_ascii_strcasecmp (value, "256"))
@@ -383,12 +413,14 @@ parse_colors_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_
     else if (!g_ascii_strcasecmp (value, "full")
              || !g_ascii_strcasecmp (value, "rgb")
              || !g_ascii_strcasecmp (value, "tc")
+             || !g_ascii_strcasecmp (value, "direct")
+             || !g_ascii_strcasecmp (value, "directcolor")
              || !g_ascii_strcasecmp (value, "truecolor"))
         options.mode = CHAFA_CANVAS_MODE_TRUECOLOR;
     else
     {
         g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
-                     "Colors must be one of [none, 2, 16, 240, 256, full].");
+                     "Colors must be one of [none, 2, 8, 16/8, 16, 240, 256, full].");
         result = FALSE;
     }
 
@@ -456,21 +488,22 @@ parse_dither_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_
     return result;
 }
 
+/* FIXME: Make it parse negative values too, and use sscanf() return values properly */
 static gboolean
-parse_font_ratio_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
+parse_fraction_or_real (const gchar *str, gdouble *real_out)
 {
-    gboolean result = TRUE;
+    gboolean success = FALSE;
     gdouble ratio = -1.0;
     gint width = -1, height = -1;
     gint o = 0;
 
-    sscanf (value, "%d/%d%n", &width, &height, &o);
+    sscanf (str, "%d/%d%n", &width, &height, &o);
     if (width < 0 || height < 0)
-        sscanf (value, "%d:%d%n", &width, &height, &o);
+        sscanf (str, "%d:%d%n", &width, &height, &o);
     if (width < 0 || height < 0)
-        sscanf (value, "%lf%n", &ratio, &o);
+        sscanf (str, "%lf%n", &ratio, &o);
 
-    if (o != 0 && value [o] != '\0')
+    if (o != 0 && str [o] != '\0')
     {
         width = -1;
         height = -1;
@@ -481,19 +514,58 @@ parse_font_ratio_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value
     if (width > 0 && height > 0)
         ratio = width / (gdouble) height;
 
+    if (ratio < 0.0)
+        goto out;
+
+    *real_out = ratio;
+    success = TRUE;
+
 out:
-    if (ratio <= 0.0)
+    return success;
+}
+
+static gboolean
+parse_font_ratio_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
+{
+    gdouble ratio = -1.0;
+    gboolean success = FALSE;
+
+    if (!parse_fraction_or_real (value, &ratio) || ratio <= 0.0)
     {
         g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
-                     "Font ratio must be specified as a real number or fraction.");
-        result = FALSE;
-    }
-    else
-    {
-        options.font_ratio = ratio;
+                     "Font ratio must be specified as a positive real number or fraction.");
+        goto out;
     }
 
-    return result;
+    options.font_ratio = ratio;
+    success = TRUE;
+
+out:
+    return success;
+}
+
+static gboolean
+parse_scale_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
+{
+    gdouble scale = -1.0;
+    gboolean success = FALSE;
+
+    if (!strcasecmp (value, "max") || !strcasecmp (value, "fill"))
+    {
+        scale = SCALE_MAX;
+    }
+    else if (!parse_fraction_or_real (value, &scale) || scale <= 0.0)
+    {
+        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                     "Scale must be specified as a positive real number or fraction.");
+        goto out;
+    }
+
+    options.scale = scale;
+    success = TRUE;
+
+out:
+    return success;
 }
 
 static gboolean
@@ -684,6 +756,113 @@ out:
     return result;
 }
 
+static void
+dump_glyph (gunichar c, const guint8 *pix, gint width, gint height, gint rowstride)
+{
+    gint x, y;
+    const guint8 *p;
+    gchar buf [16];
+    gint len;
+
+    len = g_unichar_to_utf8 (c, buf);
+    g_assert (len >= 1 && len < 16);
+    buf [len] = '\0';
+    fprintf (stdout,
+             "    {\n"
+             "        /* [%s] */\n"
+             "        CHAFA_SYMBOL_TAG_,\n"
+             "        0x%x,\n"
+             "        CHAFA_SYMBOL_OUTLINE_%s (",
+             buf,
+             c,
+             (width == 8 && height == 8) ? "8X8" :
+             (width == 16 && height == 8) ? "16X8" :
+             "STRANGE_SIZE");
+
+    for (y = 0; y < height; y++)
+    {
+        p = pix + y * rowstride;
+
+        fputs ("\n            \"", stdout);
+
+        for (x = 0; x < width; x++)
+        {
+            fputc (*p < 0x80 ? ' ' : 'X' , stdout);
+            p += 4;
+        }
+
+        fputs ("\"", stdout);
+    }
+
+    fputs (")\n    },\n", stdout);
+}
+
+/* Undocumented functionality; dumps a font file so the glyphs can be tweaked and turned
+ * into builtins. */
+static gboolean
+parse_dump_glyph_file_arg (G_GNUC_UNUSED const gchar *option_name, const gchar *value, G_GNUC_UNUSED gpointer data, GError **error)
+{
+    gboolean result = FALSE;
+    FileMapping *file_mapping;
+    FontLoader *font_loader;
+    ChafaSymbolMap *temp_map = NULL;
+    gunichar c;
+    gpointer c_bitmap;
+    gint width, height, rowstride;
+
+    options.skip_processing = TRUE;
+
+    file_mapping = file_mapping_new (value);
+    if (!file_mapping)
+    {
+        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                     "Unable to open glyph file '%s'.", value);
+        goto out;
+    }
+
+    font_loader = font_loader_new_from_mapping (file_mapping);
+    file_mapping = NULL;  /* Font loader owns it now */
+
+    if (!font_loader)
+    {
+        g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
+                     "Unable to load glyph file '%s'.", value);
+        goto out;
+    }
+
+    temp_map = chafa_symbol_map_new ();
+
+    while (font_loader_get_next_glyph (font_loader, &c, &c_bitmap, &width, &height))
+    {
+        gpointer pix;
+
+        /* By adding and then querying the glyph, we get the benefit of Chafa's
+         * internal scaling and postprocessing. */
+
+        chafa_symbol_map_add_glyph (temp_map, c,
+                                    CHAFA_PIXEL_RGBA8_PREMULTIPLIED, c_bitmap,
+                                    width, height, width * 4);
+        g_free (c_bitmap);
+
+        chafa_symbol_map_get_glyph (temp_map, c,
+                                    CHAFA_PIXEL_RGBA8_PREMULTIPLIED, &pix,
+                                    &width, &height, &rowstride);
+        dump_glyph (c, pix, width, height, rowstride);
+        g_free (pix);
+    }
+
+    font_loader_destroy (font_loader);
+
+    result = TRUE;
+
+out:
+    if (file_mapping)
+        file_mapping_destroy (file_mapping);
+    if (temp_map)
+        chafa_symbol_map_unref (temp_map);
+    return result;
+}
+
 static gboolean
 parse_boolean_token (const gchar *token, gboolean *value_out)
 {
@@ -901,7 +1080,20 @@ get_tty_size (TermSize *term_size_out)
      * aspect information for the font used. Sixel-capable terminals
      * like mlterm set these fields, but most others do not. */
 
-    if (w.ws_xpixel > 0 && w.ws_ypixel > 0)
+    if (w.ws_xpixel >= 32768 || w.ws_ypixel >= 32768)
+    {
+        /* https://github.com/hpjansson/chafa/issues/62 */
+
+        g_printerr ("%s: Terminal reports strange pixel dimensions of %dx%d.\n"
+                    "%s: Disregarding so as to avoid unreasonably large allocation.\n"
+                    "%s: This is sometimes caused by older versions of the 'fish' shell.\n"
+                    "%s: See https://github.com/hpjansson/chafa/issues/62 for details.\n",
+                    options.executable_name, (gint) w.ws_xpixel, (gint) w.ws_ypixel,
+                    options.executable_name,
+                    options.executable_name,
+                    options.executable_name);
+    }
+    else if (w.ws_xpixel > 0 && w.ws_ypixel > 0)
     {
         term_size_out->width_pixels = w.ws_xpixel;
         term_size_out->height_pixels = w.ws_ypixel;
@@ -1021,7 +1213,7 @@ parse_options (int *argc, char **argv [])
 {
     GError *error = NULL;
     GOptionContext *context;
-    gboolean result = TRUE;
+    gboolean result = FALSE;
     const GOptionEntry option_entries [] =
     {
         /* Note: The descriptive blurbs here are never shown to the user */
@@ -1039,6 +1231,7 @@ parse_options (int *argc, char **argv [])
         { "dither",      '\0', 0, G_OPTION_ARG_CALLBACK, parse_dither_arg,      "Dither", NULL },
         { "dither-grain",'\0', 0, G_OPTION_ARG_CALLBACK, parse_dither_grain_arg, "Dither grain", NULL },
         { "dither-intensity", '\0',  0, G_OPTION_ARG_DOUBLE,   &options.dither_intensity, "Dither intensity", NULL },
+        { "dump-glyph-file", '\0', 0, G_OPTION_ARG_CALLBACK, parse_dump_glyph_file_arg, "Dump glyph file", NULL },
         { "duration",    'd',  0, G_OPTION_ARG_DOUBLE,   &options.file_duration_s, "Duration", NULL },
         { "fg",          '\0', 0, G_OPTION_ARG_CALLBACK, parse_fg_color_arg,    "Foreground color of display", NULL },
         { "fg-only",     '\0', 0, G_OPTION_ARG_NONE,     &options.fg_only,      "Foreground only", NULL },
@@ -1053,6 +1246,7 @@ parse_options (int *argc, char **argv [])
         { "polite",      '\0', 0, G_OPTION_ARG_CALLBACK, parse_polite_arg,      "Polite", NULL },
         { "preprocess",  'p',  0, G_OPTION_ARG_CALLBACK, parse_preprocess_arg,  "Preprocessing", NULL },
         { "work",        'w',  0, G_OPTION_ARG_INT,      &options.work_factor,  "Work factor", NULL },
+        { "scale",       '\0', 0, G_OPTION_ARG_CALLBACK, parse_scale_arg,       "Scale", NULL },
         { "size",        's',  0, G_OPTION_ARG_CALLBACK, parse_size_arg,        "Output size", NULL },
         { "speed",       '\0', 0, G_OPTION_ARG_CALLBACK, parse_anim_speed_arg,  "Animation speed", NULL },
         { "stretch",     '\0', 0, G_OPTION_ARG_NONE,     &options.stretch,      "Stretch image to fix output dimensions", NULL },
@@ -1060,12 +1254,12 @@ parse_options (int *argc, char **argv [])
         { "threads",     '\0', 0, G_OPTION_ARG_INT,      &options.n_threads,    "Number of threads", NULL },
         { "threshold",   't',  0, G_OPTION_ARG_DOUBLE,   &options.transparency_threshold, "Transparency threshold", NULL },
         { "watch",       '\0', 0, G_OPTION_ARG_NONE,     &options.watch,        "Watch a file's contents", NULL },
+        /* Deprecated: Equivalent to --scale max */
         { "zoom",        '\0', 0, G_OPTION_ARG_NONE,     &options.zoom,         "Allow scaling up beyond one character per pixel", NULL },
         { NULL }
     };
     ChafaCanvasMode canvas_mode;
     ChafaPixelMode pixel_mode;
-    gboolean using_detected_size = FALSE;
 
     memset (&options, 0, sizeof (options));
 
@@ -1105,6 +1299,7 @@ parse_options (int *argc, char **argv [])
     options.font_ratio = -1.0;  /* Unset */
     options.margin_bottom = -1;  /* Unset */
     options.margin_right = -1;  /* Unset */
+    options.scale = -1.0;  /* Unset */
     options.work_factor = 5;
     options.optimization_level = G_MININT;  /* Unset */
     options.n_threads = -1;
@@ -1118,7 +1313,29 @@ parse_options (int *argc, char **argv [])
     if (!g_option_context_parse (context, argc, argv, &error))
     {
         g_printerr ("%s: %s\n", options.executable_name, error->message);
-        return FALSE;
+        goto out;
+    }
+
+    /* If help or version info was requested, print it and bail out */
+
+    if (options.show_help)
+    {
+        print_summary ();
+        options.skip_processing = TRUE;
+    }
+
+    if (options.show_version)
+    {
+        print_version ();
+        options.skip_processing = TRUE;
+    }
+
+    /* Some options preclude normal arg processing */
+
+    if (options.skip_processing)
+    {
+        result = TRUE;
+        goto out;
     }
 
     /* Detect terminal geometry */
@@ -1215,6 +1432,8 @@ parse_options (int *argc, char **argv [])
             options.dither_grain_height = 4;
         if (options.font_ratio <= 0.0)
             options.font_ratio = 1.0 / 2.0;
+        if (options.scale <= 0.0)
+            options.scale = 4.0;
     }
     else
     {
@@ -1228,12 +1447,14 @@ parse_options (int *argc, char **argv [])
             options.dither_grain_height = 1;
         if (options.font_ratio <= 0.0)
             options.font_ratio = 1.0 / 2.0;
+        if (options.scale <= 0.0)
+            options.scale = 1.0;
     }
 
     if (options.work_factor < 1 || options.work_factor > 9)
     {
         g_printerr ("%s: Work factor must be in the range [1-9].\n", options.executable_name);
-        return FALSE;
+        goto out;
     }
 
     if (options.transparency_threshold == G_MAXDOUBLE)
@@ -1242,14 +1463,10 @@ parse_options (int *argc, char **argv [])
     if (options.transparency_threshold < 0.0 || options.transparency_threshold > 1.0)
     {
         g_printerr ("%s: Transparency threshold %.1lf is not in the range [0.0-1.0].\n", options.executable_name, options.transparency_threshold);
-        return FALSE;
+        goto out;
     }
 
-    if (options.show_version)
-    {
-        print_version ();
-        return TRUE;
-    }
+    /* Collect filenames and validate count and correct usage of stdin */
 
     if (*argc > 1)
     {
@@ -1262,29 +1479,47 @@ parse_options (int *argc, char **argv [])
 
         options.args = g_list_append (NULL, g_strdup ("-"));
     }
-    else if (!options.show_help)
+    else
     {
         /* No arguments, no pipe, and no cry for help. */
         print_brief_summary ();
-        return FALSE;
+        goto out;
     }
 
-    /* This is outside the if-else ladder above because we want it to happen even if
-     * there are other arguments. */
-    if (options.show_help)
+    if (count_dash_strings (options.args) > 1)
     {
-        print_summary ();
-        return TRUE;
+        g_printerr ("%s: Dash '-' to pipe from standard input can be used at most once.\n",
+                    options.executable_name);
+        goto out;
     }
 
-    if (options.watch && g_list_length (options.args) != 1)
+    if (options.watch)
     {
-        g_printerr ("%s: Can only use --watch with exactly one file.\n", options.executable_name);
-        return FALSE;
+        if (g_list_length (options.args) != 1)
+        {
+            g_printerr ("%s: Can only use --watch with exactly one file.\n", options.executable_name);
+            goto out;
+        }
+
+        if (!strcmp (options.args->data, "-"))
+        {
+            g_printerr ("%s: Can only use --watch with a filename, not a pipe.\n", options.executable_name);
+            goto out;
+        }
     }
 
-    /* --stretch implies --zoom */
-    options.zoom |= options.stretch;
+    if (options.zoom)
+    {
+        g_printerr ("%s: Warning: --zoom is deprecated, use --scale max instead.\n",
+                    options.executable_name);
+        options.scale = SCALE_MAX;
+    }
+
+    /* --stretch implies --scale max (same as --zoom) */
+    if (options.stretch)
+    {
+        options.scale = SCALE_MAX;
+    }
 
     if (options.invert)
     {
@@ -1321,7 +1556,7 @@ parse_options (int *argc, char **argv [])
     {
         g_printerr ("%s: Optimization level %d is not in the range [0-9].\n",
                     options.executable_name, options.optimization_level);
-        return FALSE;
+        goto out;
     }
 
     /* Translate optimization level to flags */
@@ -1337,6 +1572,9 @@ parse_options (int *argc, char **argv [])
 
     chafa_set_n_threads (options.n_threads);
 
+    result = TRUE;
+
+out:
     g_option_context_free (context);
 
     return result;
@@ -1455,7 +1693,43 @@ build_string (ChafaPixelType pixel_type, const guint8 *pixels,
     return gs;
 }
 
-static gboolean
+static void
+pixel_to_cell_dimensions (gdouble scale,
+                          gint cell_width, gint cell_height,
+                          gint width, gint height,
+                          gint *width_out, gint *height_out)
+{
+    /* Scale can't be zero or negative */
+    scale = MAX (scale, 0.00001);
+
+    /* Zero or negative cell dimensions -> presumably unknown, use 8x8 */
+    if (cell_width < 1)
+        cell_width = 8;
+    if (cell_height < 1)
+        cell_height = 8;
+
+    if (width_out)
+    {
+        *width_out = (gdouble) width * scale / cell_width + 0.5;
+        *width_out = MAX (*width_out, 1);
+    }
+
+    if (height_out)
+    {
+        *height_out = (gdouble) height * scale / cell_width + 0.5;
+        *height_out = MAX (*height_out, 1);
+    }
+}
+
+typedef enum
+{
+    FILE_FAILED,
+    FILE_WAS_STILL,
+    FILE_WAS_ANIMATION
+}
+RunResult;
+
+static RunResult
 run_generic (const gchar *filename, gboolean is_first_file, gboolean is_first_frame, gboolean quiet)
 {
     gboolean is_animation = FALSE;
@@ -1466,6 +1740,7 @@ run_generic (const gchar *filename, gboolean is_first_file, gboolean is_first_fr
     gchar buf [CHAFA_TERM_SEQ_LENGTH_MAX * 2 + 3];
     GString *gs;
     gchar *p0;
+    RunResult result = FILE_FAILED;
 
     timer = g_timer_new ();
 
@@ -1483,6 +1758,7 @@ run_generic (const gchar *filename, gboolean is_first_file, gboolean is_first_fr
         goto out;
 
     is_animation = options.animate ? media_loader_get_is_animation (media_loader) : FALSE;
+    result = is_animation ? FILE_WAS_ANIMATION : FILE_WAS_STILL;
 
     do
     {
@@ -1501,6 +1777,7 @@ run_generic (const gchar *filename, gboolean is_first_file, gboolean is_first_fr
             gint delay_ms;
             ChafaPixelType pixel_type;
             gint src_width, src_height, src_rowstride;
+            gint virt_src_width, virt_src_height;
             gint dest_width, dest_height;
             const guint8 *pixels;
 
@@ -1519,15 +1796,31 @@ run_generic (const gchar *filename, gboolean is_first_file, gboolean is_first_fr
 
             delay_ms = media_loader_get_frame_delay (media_loader);
 
+            /* Hack to work around the fact that chafa_calc_canvas_geometry() doesn't
+             * support arbitrary scaling. Instead, we manipulate the source size to
+             * achieve the desired effect. */
+            if (using_detected_size && options.scale < SCALE_MAX - 0.1)
+            {
+                pixel_to_cell_dimensions (options.scale,
+                                          options.cell_width, options.cell_height,
+                                          src_width, src_height,
+                                          &virt_src_width, &virt_src_height);
+            }
+            else
+            {
+                virt_src_width = src_width;
+                virt_src_height = src_height;
+            }
+
             dest_width = options.width;
             dest_height = options.height;
 
-            chafa_calc_canvas_geometry (src_width,
-                                        src_height,
+            chafa_calc_canvas_geometry (virt_src_width,
+                                        virt_src_height,
                                         &dest_width,
                                         &dest_height,
                                         options.font_ratio,
-                                        options.zoom,
+                                        options.scale >= SCALE_MAX - 0.1 ? TRUE : FALSE,
                                         options.stretch);
 
             gs = build_string (pixel_type, pixels,
@@ -1621,10 +1914,10 @@ out:
         media_loader_destroy (media_loader);
     g_timer_destroy (timer);
 
-    return is_animation;
+    return result;
 }
 
-static gboolean
+static RunResult
 run (const gchar *filename, gboolean is_first_file, gboolean is_first_frame, gboolean quiet)
 {
     return run_generic (filename, is_first_file, is_first_frame, quiet);
@@ -1673,7 +1966,10 @@ static int
 run_all (GList *filenames)
 {
     GList *l;
+    gint n_processed = 0;
+    gint n_failed = 0;
 
+    /* This can only happen with --help and --version, so no error */
     if (!filenames)
         return 0;
 
@@ -1682,11 +1978,15 @@ run_all (GList *filenames)
     for (l = filenames; l && !interrupted_by_user; l = g_list_next (l))
     {
         gchar *filename = l->data;
-        gboolean was_animation;
+        RunResult result;
 
-        was_animation = run (filename, l->prev ? FALSE : TRUE, TRUE, FALSE);
+        result = run (filename, l->prev ? FALSE : TRUE, TRUE, FALSE);
 
-        if (!was_animation && options.file_duration_s != G_MAXDOUBLE)
+        n_processed++;
+        if (result == FILE_FAILED)
+            n_failed++;
+
+        if (result == FILE_WAS_STILL && options.file_duration_s != G_MAXDOUBLE)
         {
             interruptible_usleep (options.file_duration_s * 1000000.0);
         }
@@ -1697,7 +1997,7 @@ run_all (GList *filenames)
         write_to_stdout ("\n", 1);
 
     tty_options_deinit ();
-    return 0;
+    return (n_processed - n_failed < 1) ? 2 : (n_failed > 0) ? 1 : 0;
 }
 
 static void
@@ -1714,6 +2014,10 @@ proc_init (void)
     sa.sa_flags = SA_RESETHAND;
 
     sigaction (SIGINT, &sa, NULL);
+
+    /* Chafa may create and destroy GThreadPools multiple times while rendering
+     * an image. This reduces thread churn and saves a decent amount of CPU. */
+    g_thread_pool_set_max_unused_threads (-1);
 }
 
 int
@@ -1724,7 +2028,7 @@ main (int argc, char *argv [])
     proc_init ();
 
     if (!parse_options (&argc, &argv))
-        exit (1);
+        exit (2);
 
     ret = options.watch
         ? run_watch (options.args->data)
